// 9.806


#include "lighthelperPom.fx"

//-----------------------------------------------------------------------------
// Parameters
//-----------------------------------------------------------------------------
//float4x4 mWVP			: WorldViewProjection;
//float4x4 mW				: World;
//float4  EyePositionWS	: ViewPosition;

float	fHeightMapScale		= 0.04;
float	fTexelsPerSide		= sqrt(512.0 * 512.0 * 2);

int		nMaxSamples			= 50;
int		nMinSamples			= 4;

sampler Sampler
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Wrap;
    AddressV = Wrap;
};

 
 
cbuffer cbPerFrame
{
	Light gLight;
    int gLightType;
	float3 gEyePosW;
};

cbuffer cbPerObject
{
	float4x4 gWorld;
	float4x4 gWVP; 
	float4x4 gTexMtx;
};

// Nonnumeric values cannot be added to a cbuffer.
Texture2D gDiffuseMap;
Texture2D gSpecMap;

SamplerState gTriLinearSam
{
	Filter = ANISOTROPIC;
	AddressU = Wrap;
	AddressV = Wrap;
};

struct VS_IN
{
	float3 posL    : POSITION;
	float3 normalL : NORMAL;
    float4 diffuse : DIFFUSE;
    float4 spec    : SPECULAR;
	float2 texC    : TEXCOORD;
	float3 tangent   : TANGENT;
};

struct VS_OUT
{
	float4 posH    : SV_POSITION;
    float3 posW    : POSITION;
    float3 normalW : NORMAL;
    float4 diffuse : DIFFUSE;
    float4 spec    : SPECULAR;
    float2 texC    : TEXCOORD;
    
    float3 eye    : TEXCOORD1;
    float3 tangent : TEXCOORD2;
};


 
VS_OUT VS(VS_IN vIn)
{
	VS_OUT vOut;
	
	// Transform to PrimitiveObject space space.
	vOut.posW    = mul(float4(vIn.posL, 1.0f), gWorld);
	vOut.normalW = mul(float4(vIn.normalL, 0.0f), gWorld);
		
	// Transform to homogeneous clip space.
	vOut.posH = mul(float4(vIn.posL, 1.0f), gWVP);
	
	// Output vertex attributes for interpolation across triangle.
	vOut.texC  = mul(float4(vIn.texC, 0.0f, 1.0f), gTexMtx);
    
    vOut.diffuse  = vIn.diffuse;
    vOut.spec  = vIn.spec;
	
	return vOut;
}

float4 PS(VS_OUT pIn) : SV_Target
{
	// Get materials from texture maps.
	float4 diffuse = gDiffuseMap.Sample( gTriLinearSam, pIn.texC );
	float4 spec    = gSpecMap.Sample( gTriLinearSam, pIn.texC );
	
	// Interpolating normal can make it not be of unit length so normalize it.
    float3 normalW = normalize(pIn.normalW);
    
	// Compute the lit color for this pixel.
    SurfaceInfo v = {pIn.posW, normalW, diffuse, spec};
	float3 litColor = ParallelLight(v, gLight, gEyePosW);
    
    return float4(litColor, diffuse.a);
    
    //float4 test = float4( 1, 0, 0, 1 );

	//return test;
}

//Vertex shader for PARALLAX OCCLUSION MAPPING
//Based on code from http://www.gamedev.net/columns/hardcore/pom/page2.asp
VS_OUT VSpom(VS_IN vIn)
{

   VS_OUT OUT;

	float4 VertexPositionWS = mul( float4(vIn.posL, 1), gWorld );
	float3 P = VertexPositionWS.xyz;
	float3 E = gEyePosW - P;

	// The per-vertex tangent, binormal, normal form the tangent to object space
	// rotation matrix.  Multiply by the world matrix to form tangent to world space
	// rotation matrix.  Then transpose the matrix to form the inverse tangent to
	// world space, otherwise called the world to tangent space rotation matrix.
	
	float3x3 tangentToWorldSpace;									
										

																
	tangentToWorldSpace[0] = mul( vIn.tangent, gWorld );					
	tangentToWorldSpace[1] = cross(mul( vIn.normalL, gWorld ),mul( vIn.tangent, gWorld ));// SHOULD BE BINORMAL			
	tangentToWorldSpace[2] = mul( vIn.normalL, gWorld );					

	float3x3 worldToTangentSpace = transpose(tangentToWorldSpace);

	OUT.posW = mul( float4(vIn.posL, 1), gWVP );		// standard transformation to clip space
	OUT.posH = mul( float4(vIn.posL, 1), gWVP );
	OUT.texC = vIn.texC;								// pass through the original texcoords

	OUT.eye = mul( E, worldToTangentSpace );				// pass through the TANGENT space eye vector (un-normalized)
	OUT.normalW = mul( vIn.normalL, worldToTangentSpace );		// pass through the TANGENT space normal vector (un-normalized)

	return OUT;

}




//PIXEL shader for PARALLAX OCCLUSION MAPPING
//Based on code from http://www.gamedev.net/columns/hardcore/pom/page2.asp
float4 PSpom(VS_OUT pIn) : SV_Target
{
	

	float fParallaxLimit = length(pIn.eye.xy) / pIn.eye.z;	// calculate the parallax offset vector max length
	fParallaxLimit *= fHeightMapScale;						// scale the vector according to height-map scale

	float2 vOffset = normalize( -pIn.eye.xy );				// calculate the parallax offset vector direction
	vOffset = vOffset * fParallaxLimit;						// calculate the final maximum offset vector

	float3 E = normalize(pIn.eye);
	float3 N = normalize(pIn.normalW);

	//int nNumSamples = (int)lerp( nMaxSamples, nMinSamples, dot( E, N ) );	// calculate dynamic number of samples (Tatarchuk's method)
	int nNumSamples = (int)(fParallaxLimit * fTexelsPerSide);				// calculate dynamic number of samples (Zink's method)
	float fStepSize = 1.0 / (float)nNumSamples;								// calculate the texcoord step size
	
	float2 dx, dy;
	dx = ddx( pIn.texC );	// calculate the texcoord partial derivative in x in screen space for tex2Dgrad
	dy = ddy( pIn.texC );	// calculate the texcoord partial derivative in y in screen space for tex2Dgrad

	float2 vOffsetStep = fStepSize * vOffset;
	float2 vCurrOffset = float2( 0, 0 );
	float2 vLastOffset = float2( 0, 0 );
	float2 vFinalOffset = float2( 0, 0 );
	
	float4 vCurrSample;
	float4 vCurrDiffuseSample;
	float4 vLastSample;

	float stepHeight = 2.0;	
	int nCurrSample = 0;

	nNumSamples = 10; //ADDED BY TOM
	while ( nCurrSample < nNumSamples )
	{
	
		//nCurrSample++; // ADDED BY TOM
		
		//vCurrSample = tex2Dgrad( Sampler, pIn.texC + vCurrOffset, dx, dy );			// sample the current texcoord offset
		vCurrSample = gSpecMap.Sample(Sampler, pIn.texC + vCurrOffset);
		vCurrDiffuseSample = gDiffuseMap.Sample(Sampler, pIn.texC + vCurrOffset);
		vLastSample = vCurrSample;

		if ( (vCurrSample.r+vCurrSample.g+vCurrSample.b)   > stepHeight )
		{
			// calculate the linear intersection point
			float Ua = (vLastSample.a - (stepHeight+fStepSize)) / ( fStepSize + (vCurrSample.a - vLastSample.a));
			vFinalOffset = vLastOffset + Ua * vOffsetStep;

			//vCurrSample = tex2Dgrad( Sampler, pIn.texC + vFinalOffset, dx, dy );		// sample the corrected tex coords
			vCurrSample = gSpecMap.Sample(Sampler, pIn.texC + vFinalOffset);
			vCurrDiffuseSample = gDiffuseMap.Sample(Sampler, pIn.texC + vCurrOffset);
			nCurrSample = nNumSamples + 1;												// exit the while loop
		}
		else
		{
			nCurrSample++;							// increment to the next sample
			stepHeight -= fStepSize;				// change the required height-map height
			vLastOffset = vCurrOffset;				// remember this texcoord offset for next time
			vCurrOffset += vOffsetStep;				// increment to the next texcoord offset
			vLastSample = vCurrSample;
		}
	}

	//return vCurrDiffuseSample;						// return the final sampled color value
	
	
	//now do some lightning
	
	// Get materials from texture maps.
	float4 diffuse = vCurrDiffuseSample;
	float4 spec    = float4(0,0,0,1);//gSpecMap.Sample( gTriLinearSam, pIn.texC );
	
	// Interpolating normal can make it not be of unit length so normalize it.
    float3 normalW = normalize(pIn.normalW);
    
	// Compute the lit color for this pixel.
    SurfaceInfo v = {pIn.posW, normalW, diffuse, spec};
	float3 litColor = ParallelLight(v, gLight, gEyePosW);
    
    return float4(litColor, diffuse.a);
}



technique10 TexTech
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, VS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS() ) );
    }
}

technique10 parallaxocclusionmapping
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, VSpom() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PSpom() ) );
    }
}
